@@ -5,14 +5,17 @@ import logging
 import os
 import pickle
 import tensorflow as tf
-from typing import Callable, Dict, List, Union
+from tensorflow_probability.python.distributions.distribution import Distribution
+from typing import Callable, Dict, List, Tuple, Union
 from alibi_detect.ad import AdversarialAE
 from alibi_detect.ad.adversarialae import DenseHidden
 from alibi_detect.base import BaseDetector
 from alibi_detect.cd import KSDrift, MMDDrift
 from alibi_detect.models.autoencoder import AE, AEGMM, DecoderLSTM, EncoderLSTM, Seq2Seq, VAE, VAEGMM
-from alibi_detect.od import (IForest, Mahalanobis, OutlierAE, OutlierAEGMM, OutlierProphet,
+from alibi_detect.models.pixelcnn import PixelCNN
+from alibi_detect.od import (IForest, LLR, Mahalanobis, OutlierAE, OutlierAEGMM, OutlierProphet,
                              OutlierSeq2Seq, OutlierVAE, OutlierVAEGMM, SpectralResidual)
+from alibi_detect.od.llr import build_model
 
 logger = logging.getLogger(__name__)
 
@@ -21,6 +24,7 @@ Data = Union[
     AdversarialAE,
     IForest,
     KSDrift,
+    LLR,
     Mahalanobis,
     MMDDrift,
     OutlierAEGMM,
@@ -36,6 +40,7 @@ DEFAULT_DETECTORS = [
     'AdversarialAE',
     'IForest',
     'KSDrift',
+    'LLR',
     'Mahalanobis',
     'MMDDrift',
     'OutlierAE',
@@ -51,7 +56,7 @@ DEFAULT_DETECTORS = [
 def save_detector(detector: Data,
                   filepath: str) -> None:
     """
-    Save outlier or adversarial detector.
+    Save outlier, drift or adversarial detector.
 
     Parameters
     ----------
@@ -97,6 +102,8 @@ def save_detector(detector: Data,
         state_dict = state_sr(detector)
     elif detector_name == 'OutlierSeq2Seq':
         state_dict = state_s2s(detector)
+    elif detector_name == 'LLR':
+        state_dict = state_llr(detector)
 
     with open(os.path.join(filepath, detector_name + '.pickle'), 'wb') as f:
         pickle.dump(state_dict, f)
@@ -116,32 +123,8 @@ def save_detector(detector: Data,
         save_tf_hl(detector.model_hl, filepath)
     elif detector_name == 'OutlierSeq2Seq':
         save_tf_s2s(detector, filepath)
-    elif detector_name in ['KSDrift', 'MMDDrift']:
-        save_tf_preprocess(detector, filepath)
-
-
-def state_preprocess(state_dict: dict, cd: Union[KSDrift, MMDDrift]) -> dict:
-    """
-    State of preprocessing functions to save.
-
-    Parameters
-    ----------
-    state_dict
-        State dict for detector.
-    cd
-        Drift detection object
-    """
-    if isinstance(cd.preprocess_fn, Callable):
-        if cd.preprocess_fn.__name__ in ['uae', 'hidden_output']:
-            kwargs_list = list(cd.preprocess_kwargs.keys())
-            k = None
-            if 'model' in kwargs_list:
-                k = 'model'
-            if 'encoder_net' in kwargs_list:
-                k = 'encoder_net'
-            if isinstance(k, str):
-                state_dict['preprocess_kwargs'][k] = k
-    return state_dict
+    elif detector_name == 'LLR':
+        save_tf_llr(detector, filepath)
 
 
 def state_ksdrift(cd: KSDrift) -> Dict:
@@ -156,15 +139,14 @@ def state_ksdrift(cd: KSDrift) -> Dict:
     state_dict = {
         'p_val': cd.p_val,
         'X_ref': cd.X_ref,
+        'preprocess_X_ref': cd.preprocess_X_ref,
         'update_X_ref': cd.update_X_ref,
         'alternative': cd.alternative,
         'n': cd.n,
         'n_features': cd.n_features,
         'correction': cd.correction,
-        'preprocess_fn': cd.preprocess_fn,
-        'preprocess_kwargs': cd.preprocess_kwargs.copy() if isinstance(cd.preprocess_kwargs, dict) else None
     }
-    return state_preprocess(state_dict, cd)
+    return state_dict
 
 
 def state_mmddrift(cd: MMDDrift) -> Dict:
@@ -179,15 +161,14 @@ def state_mmddrift(cd: MMDDrift) -> Dict:
     state_dict = {
         'p_val': cd.p_val,
         'X_ref': cd.X_ref,
+        'preprocess_X_ref': cd.preprocess_X_ref,
         'update_X_ref': cd.update_X_ref,
         'n': cd.n,
         'chunk_size': cd.chunk_size,
         'permutation_test': cd.permutation_test,
         'infer_sigma': cd.infer_sigma,
-        'preprocess_fn': cd.preprocess_fn,
-        'preprocess_kwargs': cd.preprocess_kwargs.copy() if isinstance(cd.preprocess_kwargs, dict) else None
     }
-    return state_preprocess(state_dict, cd)
+    return state_dict
 
 
 def state_iforest(od: IForest) -> Dict:
@@ -371,6 +352,24 @@ def state_s2s(od: OutlierSeq2Seq) -> Dict:
     return state_dict
 
 
+def state_llr(od: LLR) -> Dict:
+    """
+    LLR parameters to save.
+
+    Parameters
+    ----------
+    od
+        Outlier detector object.
+    """
+    state_dict = {
+        'threshold': od.threshold,
+        'has_log_prob': od.has_log_prob,
+        'sequential': od.sequential,
+        'log_prob': od.log_prob
+    }
+    return state_dict
+
+
 def save_tf_ae(detector: Union[OutlierAE, AdversarialAE],
                filepath: str) -> None:
     """
@@ -475,6 +474,34 @@ def save_tf_model(model: tf.keras.Model,
         logger.warning('No `tf.keras.Model` or `tf.keras.Sequential` detected. No model saved.')
 
 
+def save_tf_llr(detector: LLR, filepath: str) -> None:
+    """
+    Save LLR TensorFlow models or distributions.
+
+    Parameters
+    ----------
+    detector
+        Outlier detector object.
+    filepath
+        Save directory.
+    """
+    if not os.path.isdir(filepath):
+        logger.warning('Directory {} does not exist and is now created.'.format(filepath))
+        os.mkdir(filepath)
+    model_dir = os.path.join(filepath, 'model')
+    if not os.path.isdir(model_dir):
+        os.mkdir(model_dir)
+    if hasattr(detector, 'model_s') and hasattr(detector, 'model_b'):
+        if not os.path.isdir(model_dir):
+            os.mkdir(model_dir)
+        detector.model_s.save_weights(os.path.join(model_dir, 'model_s.h5'))
+        detector.model_b.save_weights(os.path.join(model_dir, 'model_b.h5'))
+    else:
+        detector.dist_s.save(os.path.join(model_dir, 'model.h5'))
+        if detector.dist_b is not None:
+            detector.dist_b.save(os.path.join(model_dir, 'model_background.h5'))
+
+
 def save_tf_hl(models: List[tf.keras.Model],
                filepath: str) -> None:
     """
@@ -608,22 +635,9 @@ def save_tf_s2s(od: OutlierSeq2Seq,
         logger.warning('No `tf.keras.Model` Seq2Seq detected. No Seq2Seq model saved.')
 
 
-def save_tf_preprocess(cd: Union[KSDrift, MMDDrift], filepath: str) -> None:
-    if isinstance(cd.preprocess_fn, Callable):
-        if cd.preprocess_fn.__name__ in ['uae', 'hidden_output']:
-            kwargs_list = list(cd.preprocess_kwargs.keys())
-            k = None
-            if 'model' in kwargs_list:
-                k = 'model'
-            elif 'encoder_net' in kwargs_list:
-                k = 'encoder_net'
-            if isinstance(k, str):
-                save_tf_model(cd.preprocess_kwargs[k], filepath, model_name=k)
-
-
 def load_detector(filepath: str, **kwargs) -> Data:
     """
-    Load outlier or adversarial detector.
+    Load outlier, drift or adversarial detector.
 
     Parameters
     ----------
@@ -684,11 +698,12 @@ def load_detector(filepath: str, **kwargs) -> Data:
         seq2seq = load_tf_s2s(filepath, state_dict)
         detector = init_od_s2s(state_dict, seq2seq)
     elif detector_name == 'KSDrift':
-        model = load_tf_preprocess(filepath, state_dict)
-        detector = init_cd_ksdrift(state_dict, model)
+        detector = init_cd_ksdrift(state_dict, **kwargs)
     elif detector_name == 'MMDDrift':
-        model = load_tf_preprocess(filepath, state_dict)
-        detector = init_cd_mmddrift(state_dict, model)
+        detector = init_cd_mmddrift(state_dict, **kwargs)
+    elif detector_name == 'LLR':
+        models = load_tf_llr(filepath, **kwargs)
+        detector = init_od_llr(state_dict, models)
 
     detector.meta = meta_dict
     return detector
@@ -881,34 +896,40 @@ def load_tf_s2s(filepath: str,
     return seq2seq
 
 
-def load_tf_preprocess(filepath: str, state_dict: dict) -> Union[tf.keras.Model, tf.keras.Sequential]:
+def load_tf_llr(filepath: str, dist_s: Union[Distribution, PixelCNN] = None,
+                dist_b: Union[Distribution, PixelCNN] = None, input_shape: tuple = None):
     """
-    Load TensorFlow model for KSDrift or MMDDrift detector.
+    Load LLR TensorFlow models or distributions.
 
     Parameters
     ----------
+    detector
+        Likelihood ratio detector.
     filepath
-        Saved model directory.
-    state_dict
-        Dictionary containing the detector's parameters.
+        Save directory.
+    dist_s
+        TensorFlow distribution for semantic model.
+    dist_b
+        TensorFlow distribution for background model.
+    input_shape
+        Input shape of the model.
 
     Returns
     -------
-    Loaded model.
+    Detector with loaded models.
     """
-    model = None
-    preprocess_fn = state_dict['preprocess_fn']
-    if isinstance(preprocess_fn, Callable):
-        if preprocess_fn.__name__ in ['uae', 'hidden_output']:
-            kwargs_list = list(state_dict['preprocess_kwargs'])
-            k = None
-            if 'model' in kwargs_list:
-                k = 'model'
-            elif 'encoder_net' in kwargs_list:
-                k = 'encoder_net'
-            if isinstance(k, str):
-                model = load_tf_model(filepath, model_name=k)
-    return model
+    model_dir = os.path.join(filepath, 'model')
+    if 'model_s.h5' in os.listdir(model_dir) and 'model_b.h5' in os.listdir(model_dir):
+        model_s, dist_s = build_model(dist_s, input_shape, os.path.join(model_dir, 'model_s.h5'))
+        model_b, dist_b = build_model(dist_b, input_shape, os.path.join(model_dir, 'model_b.h5'))
+        return dist_s, dist_b, model_s, model_b
+    else:
+        dist_s = tf.keras.models.load_model(os.path.join(model_dir, 'model.h5'), compile=False)
+        if 'model_background.h5' in os.listdir(model_dir):
+            dist_b = tf.keras.models.load_model(os.path.join(model_dir, 'model_background.h5'), compile=False)
+        else:
+            dist_b = None
+        return dist_s, dist_b, None, None
 
 
 def init_od_ae(state_dict: Dict,
@@ -1073,7 +1094,18 @@ def init_od_s2s(state_dict: Dict,
     return od
 
 
-def init_cd_ksdrift(state_dict: Dict, model: Union[tf.keras.Model, tf.keras.Sequential] = None) -> KSDrift:
+def init_preprocess(**kwargs) -> Tuple[Callable, dict]:
+    """ Return preprocessing function and kwargs. """
+    if isinstance(kwargs, dict):
+        keys = list(kwargs.keys())
+        preprocess_fn = kwargs['preprocess_fn'] if 'preprocess_fn' in keys else None
+        preprocess_kwargs = kwargs['preprocess_kwargs'] if 'preprocess_kwargs' in keys else None
+    else:
+        preprocess_fn, preprocess_kwargs = None, None
+    return preprocess_fn, preprocess_kwargs
+
+
+def init_cd_ksdrift(state_dict: Dict, **kwargs) -> KSDrift:
     """
     Initialize KSDrift detector.
 
@@ -1081,24 +1113,18 @@ def init_cd_ksdrift(state_dict: Dict, model: Union[tf.keras.Model, tf.keras.Sequ
     ----------
     state_dict
         Dictionary containing the parameter values.
-    model
-        Optionally loaded model.
+    kwargs
+        Kwargs optionally containing preprocess_fn and preprocess_kwargs.
 
     Returns
     -------
     Initialized KSDrift instance.
     """
-    preprocess_fn = state_dict['preprocess_fn']
-    preprocess_kwargs = state_dict['preprocess_kwargs']
-    if isinstance(preprocess_fn, Callable) and isinstance(model, (tf.keras.Model, tf.keras.Sequential)):
-        if preprocess_fn.__name__ == 'uae':
-            preprocess_kwargs['encoder_net'] = model
-        elif preprocess_fn.__name__ == 'hidden_output':
-            preprocess_kwargs['model'] = model
-
+    preprocess_fn, preprocess_kwargs = init_preprocess(**kwargs)
     cd = KSDrift(
         p_val=state_dict['p_val'],
         X_ref=state_dict['X_ref'],
+        preprocess_X_ref=False,
         update_X_ref=state_dict['update_X_ref'],
         preprocess_fn=preprocess_fn,
         preprocess_kwargs=preprocess_kwargs,
@@ -1107,10 +1133,11 @@ def init_cd_ksdrift(state_dict: Dict, model: Union[tf.keras.Model, tf.keras.Sequ
         n_features=state_dict['n_features']
     )
     cd.n = state_dict['n']
+    cd.preprocess_X_ref = state_dict['preprocess_X_ref']
     return cd
 
 
-def init_cd_mmddrift(state_dict: Dict, model: Union[tf.keras.Model, tf.keras.Sequential] = None) -> MMDDrift:
+def init_cd_mmddrift(state_dict: Dict, **kwargs) -> MMDDrift:
     """
     Initialize MMDDrift detector.
 
@@ -1118,30 +1145,25 @@ def init_cd_mmddrift(state_dict: Dict, model: Union[tf.keras.Model, tf.keras.Seq
     ----------
     state_dict
         Dictionary containing the parameter values.
-    model
-        Optionally loaded model.
+    kwargs
+        Kwargs optionally containing preprocess_fn and preprocess_kwargs.
 
     Returns
     -------
     Initialized MMDDrift instance.
     """
-    preprocess_fn = state_dict['preprocess_fn']
-    preprocess_kwargs = state_dict['preprocess_kwargs']
-    if isinstance(preprocess_fn, Callable) and isinstance(model, (tf.keras.Model, tf.keras.Sequential)):
-        if preprocess_fn.__name__ == 'uae':
-            preprocess_kwargs['encoder_net'] = model
-        elif preprocess_fn.__name__ == 'hidden_output':
-            preprocess_kwargs['model'] = model
-
+    preprocess_fn, preprocess_kwargs = init_preprocess(**kwargs)
     cd = MMDDrift(
         p_val=state_dict['p_val'],
         X_ref=state_dict['X_ref'],
+        preprocess_X_ref=False,
         update_X_ref=state_dict['update_X_ref'],
         preprocess_fn=preprocess_fn,
         preprocess_kwargs=preprocess_kwargs,
         chunk_size=state_dict['chunk_size']
     )
     cd.n = state_dict['n']
+    cd.preprocess_X_ref = state_dict['preprocess_X_ref']
     cd.infer_sigma = state_dict['infer_sigma']
     cd.permutation_test = state_dict['permutation_test']
     return cd
@@ -1230,3 +1252,27 @@ def init_od_sr(state_dict: Dict) -> SpectralResidual:
                           n_est_points=state_dict['n_est_points'],
                           n_grad_points=state_dict['n_grad_points'])
     return od
+
+
+def init_od_llr(state_dict: Dict, models: tuple) -> LLR:
+    """
+    Initialize LLR detector.
+
+    Parameters
+    ----------
+    state_dict
+        Dictionary containing the parameter values.
+
+    Returns
+    -------
+    Initialized LLR instance.
+    """
+    od = LLR(threshold=state_dict['threshold'],
+             model=models[0],
+             model_background=models[1],
+             log_prob=state_dict['log_prob'],
+             sequential=state_dict['sequential'])
+    if models[2] is not None and models[3] is not None:
+        od.model_s = models[2]
+        od.model_b = models[3]
+    return od
